@@ -23,20 +23,23 @@ from cryptography.hazmat.backends import default_backend
 
 # Own modules
 from pygnmi.path_generator import gnmi_path_generator, gnmi_path_degenerator
+from pygnmi.tools import diff_openconfig
 
 
 # Logger
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.StreamHandler(sys.stdout))
 
+
 # Classes
 class gNMIclient(object):
     """
     This class instantiates the object, which interacts with the network elements over gNMI.
     """
     def __init__(self, target: tuple, username: str = None, password: str = None,
-                 debug: bool = False, insecure: bool = False, path_cert: str = None, path_key: str = None, path_root: str = None, override: str = None,
-                 gnmi_timeout: int = 5, grpc_options: list = [], **kwargs ):
+                 debug: bool = False, insecure: bool = False, path_cert: str = None, 
+                 path_key: str = None, path_root: str = None, override: str = None,
+                 gnmi_timeout: int = 5, grpc_options: list = [], show_diff: str = "", **kwargs ):
 
         """
         Initializing the object
@@ -50,6 +53,7 @@ class gNMIclient(object):
         self.__path_root = path_root
         self.__options=([('grpc.ssl_target_name_override', override)]+grpc_options) if override else grpc_options
         self.__gnmi_timeout = gnmi_timeout
+        self.__show_diff = show_diff if show_diff in {"get", "print"} else ""
 
         self.__target_path = f'{target[0]}:{target[1]}'
         if re.match('unix:.*', target[0]):
@@ -88,9 +92,10 @@ class gNMIclient(object):
         return self.connect()
 
 
-    def connect(self):
+    def connect(self,timeout:int = None):
         """
         Building the connectivity towards network element over gNMI
+        timeout: optional override of the time to wait for connection, defaults to init parameter
         """
 
         if self.__insecure:
@@ -133,18 +138,22 @@ class gNMIclient(object):
             self.__channel = grpc.secure_channel(self.__target_path,
                                                  credentials=cert, options=self.__options)
 
-        if self.__gnmi_timeout is None or self.__gnmi_timeout > 0:
-           self.wait_for_connect(self.__gnmi_timeout)
+        if timeout is None:
+           timeout = self.__gnmi_timeout
+        if timeout is None or timeout > 0:
+           self.wait_for_connect(timeout)
         self.__stub = gNMIStub(self.__channel)
 
         return self
 
+
     def wait_for_connect(self, timeout: int):
         """
         Wait for the gNMI connection to the server to come up, with given timeout
         """
         grpc.channel_ready_future(self.__channel).result(timeout=timeout)
 
+
     def capabilities(self):
         """
         Collecting the gNMI capabilities of the network device.
@@ -256,7 +265,7 @@ class gNMIclient(object):
             elif datatype == 'operational':
                 pb_datatype = 3
             else:
-                logger.error('The GetRequst data type is not within the dfined range')
+                logger.error('The GetRequst data type is not within the defined range')
 
         if encoding in encoding_set:
             if encoding.lower() == 'json':
@@ -315,6 +324,7 @@ class gNMIclient(object):
             if gnmi_message_response:
                 response = {}
 
+                ## Message GetRespone, Key notification
                 if gnmi_message_response.notification:
                     response.update({'notification': []})
 
@@ -327,6 +337,12 @@ class gNMIclient(object):
                         ## Message Notification, Key prefix
                         notification_container.update({'prefix': gnmi_path_degenerator(notification.prefix)}) if notification.prefix else notification_container.update({'prefix': None})
 
+                        ## Message Notification, Key alias
+                        notification_container.update({'alias': notification.alias}) if notification.alias else notification_container.update({'alias': None})
+
+                        ## Message Notification, Key atomic
+                        notification_container.update({'atomic': notification.atomic })
+
                         ## Message Notification, Key update
                         if notification.update:
                             notification_container.update({'update': []})
@@ -335,7 +351,7 @@ class gNMIclient(object):
                                 update_container = {}
 
                                 ## Message Update, Key path
-                                update_container.update({'path': gnmi_path_degenerator(update_msg.path )}) if update_msg.path else update_container.update({'path': None})
+                                update_container.update({'path': gnmi_path_degenerator(update_msg.path)}) if update_msg.path else update_container.update({'path': None})
 
                                 ## Message Update, Key val
                                 if update_msg.HasField('val'):
@@ -417,6 +433,7 @@ class gNMIclient(object):
         replace_msg = []
         update_msg = []
         encoding_set = {'json', 'bytes', 'proto', 'ascii', 'json_ietf'}
+        diff_list = []
 
         if encoding not in encoding_set:
             logger.error(f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
@@ -488,6 +505,16 @@ class gNMIclient(object):
                 raise Exception ('The provided input for Set message (replace operation) is not list.')
 
         try:
+            ## Adding collection of data for diff before the change
+            if self.__show_diff:
+                paths_to_collect_list = []
+
+                if delete: paths_to_collect_list.extend(delete)
+                if update: paths_to_collect_list.extend([path_tuple[0] for path_tuple in update])
+                if replace: paths_to_collect_list.extend([path_tuple[0] for path_tuple in replace])
+
+                pre_change_dict = self.get(path=paths_to_collect_list)
+
             gnmi_message_request = SetRequest(delete=del_protobuf_paths, update=update_msg, replace=replace_msg)
 
             if self.__debug:
@@ -505,30 +532,22 @@ class gNMIclient(object):
             if gnmi_message_response:
                 response = {}
 
+                ## Message SetResponse, Key timestamp
+                response.update({'timestamp': gnmi_message_response.timestamp}) if gnmi_message_response.timestamp else response.update({'timestamp': 0})
+
+                ## Message SetResponse, Key prefix
+                response.update({'prefix': gnmi_path_degenerator(gnmi_message_response.prefix)}) if gnmi_message_response.prefix else response.update({'prefix': None})
+
                 if gnmi_message_response.response:
                     response.update({'response': []})
 
                     for response_entry in gnmi_message_response.response:
                         response_container = {}
 
-                        if response_entry.path and response_entry.path.elem:
-                            resource_path = []
-                            for path_elem in response_entry.path.elem:
-                                tp = ''
-                                if path_elem.name:
-                                    tp += path_elem.name
-
-                                if path_elem.key:
-                                    for pk_name, pk_value in sorted(path_elem.key.items()):
-                                        tp += f'[{pk_name}={pk_value}]'
-
-                                resource_path.append(tp)
-
-                            response_container.update({'path': '/'.join(resource_path)})
-
-                        else:
-                            response_container.update({'path': None})
+                        ## Message UpdateResult, Key path
+                        response_container.update({'path': gnmi_path_degenerator(response_entry.path)}) if response_entry.path else response_container.update({'path': None})
 
+                        ## Message UpdateResult, Key op
                         if response_entry.op:
                             if response_entry.op == 1:
                                 res_op = 'DELETE'
@@ -543,7 +562,21 @@ class gNMIclient(object):
 
                         response['response'].append(response_container)
 
-                return response
+                ## Adding collection of data for diff before the change
+                if self.__show_diff:
+                    post_change_dict = self.get(path=paths_to_collect_list)
+
+                    is_printable = True if self.__show_diff == "print" else False
+
+                    diff_list = diff_openconfig(pre_dict=pre_change_dict,
+                                                post_dict=post_change_dict,
+                                                is_printable=is_printable)
+
+                if diff_list and self.__show_diff == "get":
+                    return response, diff_list
+
+                else:
+                    return response
 
             else:
                 logger.error('Failed parsing the SetResponse.')
@@ -555,9 +588,9 @@ class gNMIclient(object):
 
             raise Exception (err)
 
-        except:
-            logger.error(f'Collection of Set information failed is failed.')
-            return None
+        # except:
+        #     logger.error(f'Collection of Set information failed is failed.')
+        #     return None
 
 
     def set_with_retry(self, delete: list = None, replace: list = None, update: list = None, encoding: str = 'json', retry_delay: int = 3):
